/*
rss.c - RSS (and Atom) parser and generator (using libxml2)

Copyright (c) 2006 NoisyB


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include "xml.h"
#include "rss.h"


#ifndef _WIN32
#define stricmp strcasecmp
#define strnicmp strncasecmp
#endif


#define RSS_V0_91_S "RSS v0.91"
#define RSS_V0_92_S "RSS v0.92"
#define RSS_V0_93_S "RSS v0.93"
#define RSS_V0_94_S "RSS v0.94"
#define RSS_V1_0_S  "RSS v1.0"
#define RSS_V2_0_S  "RSS v2.0"
#define ATOM_V0_1_S "ATOM v0.1"
#define ATOM_V0_2_S "ATOM v0.2"
#define ATOM_V0_3_S "ATOM v0.3"


//#ifdef  DEBUG
void
rss_st_rss_t_sanity_check (st_rss_t *rss)
{
  int i = 0;

  for (; i < rss->item_count; i++)
    printf ("pos:%d\n"
            "title: %s\n"
            "url: %s\n"
            "date: %ld\n"
            "desc: %s\n\n",
      i,
      rss->item[i].title,
      rss->item[i].url,
      rss->item[i].date,
      rss->item[i].desc);

  printf ("rss->item_count: %d\n\n", rss->item_count);
}
//#endif


#if 0
char *
rss_utf8_enc (const char *in, const char *encoding)
{
  static xmlChar out[RSSMAXBUFSIZE * 2];
  int temp = 0;
  int size = 0;
  xmlCharEncodingHandlerPtr handler;

  if (!in)
    return NULL;

  handler = xmlFindCharEncodingHandler (encoding);

  if (!handler)
    return NULL;

  temp = strlen (in);
  size = sizeof (out);

  handler->input (out, &size, in, &temp);

  return (char *) out;
}
#endif


typedef struct
{
  int version;
  const char *version_s;

  const char *magic_s;
} st_rss_version_t;


static st_rss_version_t rss_version[] = {
  {RSS_V0_91, RSS_V0_91_S, "0.91"},
  {RSS_V0_92, RSS_V0_92_S, "0.92"},
  {RSS_V0_93, RSS_V0_93_S, "0.93"},
  {RSS_V0_94, RSS_V0_94_S, "0.94"},
  {RSS_V1_0,  RSS_V1_0_S,  "1.0"},
  {RSS_V2_0,  RSS_V2_0_S,  "2.0"},
  {RSS_V2_0,  RSS_V2_0_S,  "2"},
  {ATOM_V0_1, ATOM_V0_1_S, "0.1"},
  {ATOM_V0_2, ATOM_V0_2_S, "0.2"},
  {ATOM_V0_3, ATOM_V0_3_S, "0.3"},
  {0, NULL, NULL}
};


const char *
rss_get_version_s_by_magic (const char *m)
{
  int i = 0;

  for (; rss_version[i].version; i++)
    if (!strcmp (rss_version[i].magic_s, m))
      return rss_version[i].version_s;
  return NULL;
}


const char *
rss_get_version_s_by_id (int version)
{
  int i = 0;

  for (; rss_version[i].version; i++)
    if (rss_version[i].version == version)
      return rss_version[i].version_s;
  return NULL;
}

const char *
rss_get_version_s (st_rss_t * rss)
{
  return rss_get_version_s_by_id (rss->version);
}


unsigned int
rss_item_count (st_rss_t * rss)
{
  return rss->item_count;
}


st_rss_item_t *
rss_get_item (st_rss_t * rss, unsigned int n)
{
  return &rss->item[n];
}


#if 0
void
rsstool_set_version (const char *version)
{
  const st_rss_version_t *v = NULL;

  if (!(*version))
    fprintf (stderr, "WARNING: no version in %d\n", rsstool.article_count);

  v = rsstool_get_rss_version_by_magic (version);
  if (v)
    rsstool.rss[rsstool.article_count].version = v->version;

#ifdef  DEBUG
  printf ("link[%d]: %s\n", rsstool.article_count, rsstool.rss[rsstool.article_count].version);
  fflush (stdout);
#endif
}


void
rsstool_set_site (const char *site)
{
  if (!(*site))
    fprintf (stderr, "WARNING: no site discovered\n");

  strncpy (rsstool.rss[rsstool.article_count].site, site, RSSMAXBUFSIZE)[RSSMAXBUFSIZE - 1] = 0;

#ifdef  DEBUG
  printf ("site[%d]: %s\n", rsstool.article_count, rsstool.rss[rsstool.article_count].site);
  fflush (stdout);
#endif
}
#endif


time_t
parse_date (const char *s)
{
  int i = 0;
  char y[100], m[100], d[100];
  char h[100], min[100];
//  char sec[100];
  struct tm time_tag;
  time_t t = time (0);
  const char *month_s[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec", NULL};

  *y = *m = *d = *h = *min = 0;

  if (s[10] == 'T')                     // YYYY-MM-DDT00:00+00:00
    {
      sscanf (s, " %4s-%2s-%2sT%2s:%2s", y, m, d, h, min);
    }
  else if (s[3] == ',' && s[4] == ' ')  // Mon, 31 Jul 2006 15:05:00 GMT
    {
      sscanf (s + 5, "%2s %s %4s %2s:%2s", d, m, y, h, min);

      for (i = 0; month_s[i]; i++)
        if (!stricmp (m, month_s[i]))
          {
            sprintf (m, "%d", i + 1);
            break;
          }
    }
  else if (s[4] == '-' && s[7] == '-')  // 2006-07-19
    {
      sscanf (s, "%4s-%2s-%2s", y, m, d);
    }
  else                                  // YYYYMMDDTHHMMSS
    {
//      sscanf (s, " %4s%2s%2sT", y, m, d);
    }

  memset (&time_tag, 0, sizeof (struct tm));

  if (*y)
    time_tag.tm_year = strtol (y, NULL, 10) - 1900;
  if (*m)
    time_tag.tm_mon = strtol (m, NULL, 10) - 1;
  if (*d)
    time_tag.tm_mday = strtol (d, NULL, 10);
  if (*h)
    time_tag.tm_hour = strtol (h, NULL, 10);
  if (*min)
    time_tag.tm_min = strtol (min, NULL, 10);

  t = mktime (&time_tag);

  return t;
}


int
rss_demux (const char *fname)
{
  xml_doc_t *doc = NULL;
  xml_node_t *cur = NULL;
  int version = -1;
  char *p = NULL;

  if (!(doc = xml_parse (fname)))
    {
      fprintf (stderr, "ERROR: cannot read %s\n", fname);
      return -1;
    }

  cur = xml_get_rootnode (doc);

  if (!cur)
    return -1;

  if (!cur->name)
    return -1;

#ifdef  DEBUG
  printf ("%s\n", cur->name);
  fflush (stdout);
#endif

  if (!xml_strcmp (cur->name, "html")) // not xml
    return -1;

  if (!xml_strcmp (cur->name, "feed")) // atom
    {
      version = ATOM_V0_1; // default

      if (!(p = (char *) xml_get_prop (cur, "version")))
        return version;

      if (!strcmp (p, "0.3"))
        version = ATOM_V0_3;
      else if (!strcmp (p, "0.2"))
        version = ATOM_V0_2;
//      else if (!strcmp (p, "0.1"))
//        version = ATOM_V0_1;

      return version;
    }
  else if (!xml_strcmp (cur->name, "rss")) // rss
    {
      if (!(p = (char *) xml_get_prop (cur, "version")))
        return -1;

      if (!strcmp (p, "0.91"))
        version = RSS_V0_91;
      else if (!strcmp (p, "0.92"))
        version = RSS_V0_92;
      else if (!strcmp (p, "0.93"))
        version = RSS_V0_93;
      else if (!strcmp (p, "0.94"))
        version = RSS_V0_94;
      else if (!strcmp (p, "2") || !strcmp (p, "2.0") || !strcmp (p, "2.00"))
        version = RSS_V2_0;

      return version;
    }
  else if (!xml_strcmp (cur->name, "rdf") ||
           !xml_strcmp (cur->name, "RDF")) // rdf
    {
#if 0
      if (!(p = xml_get_prop (cur, "xmlns")))
        return -1;

      // hack
      if (!strcmp (p, "http://my.netscape.com/rdf/simple/0.9/"))
        version = RSS_V0_90;
      else if (!strcmp (p, "http://purl.org/rss/1.0/"))
        version = RSS_V1_0;
#else
      version = RSS_V1_0;
#endif

      return version;
    }

  return -1;
}


static void
rss_read_copy (char *d, xml_doc_t* doc, xml_node_t* n)
{
  (void) doc;
  const char *p = (const char *) xml_get_string (n);

  if (p)
    strncpy (d, p, RSSMAXBUFSIZE)[RSSMAXBUFSIZE-1] = 0;
  else
    *d = 0;
}



static st_rss_t *
rss_open_rss (st_rss_t *rss)
{
  xml_doc_t *doc;
  xml_node_t *cur;
  int rdf = 0;

  doc = xml_parse (rss->url);
  if (!doc)
    {
      fprintf (stderr, "ERROR: cannot read %s\n", rss->url);
      return NULL;
    }

  cur = xml_get_rootnode (doc);
  if (!cur)
    {
      fprintf (stderr, "ERROR: empty document %s\n", rss->url);
      xml_free (doc);
      return NULL;
    }

  // rdf?
  // TODO: move this to rss_demux()
  if (xml_strcmp (cur->name, "rss") != 0 &&
      (!xml_strcmp (cur->name, "rdf") ||
       !xml_strcmp (cur->name, "RDF")))
    rdf = 1;

  cur = xml_get_childnode (cur);
  while (cur && xml_is_empty_node (cur))
    cur = xml_get_nextnode (cur);
  if (!cur)
    {
//      fprintf (stderr, "");
      return NULL;
    }

  if (xml_strcmp (cur->name, "channel"))
    {
      fprintf (stderr, "ERROR: bad document: did not immediately find the RSS element\n");
      return NULL;
    }

  if (!rdf) // document is RSS
    cur = xml_get_childnode (cur);

  while (cur)
    {
      while (cur && xml_is_empty_node (cur))
        cur = xml_get_nextnode (cur);

      if (!cur)
        break;

      if (!xml_strcmp (cur->name, "title"))
        rss_read_copy (rss->title, doc, cur->xmlChildrenNode);
      else if (!xml_strcmp (cur->name, "description"))
        rss_read_copy (rss->desc, doc, cur->xmlChildrenNode);
//      else if (!xml_strcmp (cur->name, "link"))
//        rss_read_copy (rss->url, doc, cur->xmlChildrenNode);
      else if (!xml_strcmp (cur->name, "date") ||
               !xml_strcmp (cur->name, "pubDate") ||
               !xml_strcmp (cur->name, "dc:date"))
        rss->date = parse_date ((const char *) xml_get_string (cur->xmlChildrenNode));
      else if (!xml_strcmp (cur->name, "channel") && rdf)
        {
          xml_node_t *pcur = xml_get_childnode (cur);

          while (pcur)
            {
              if (!xml_strcmp (pcur->name, "title"))
                rss_read_copy (rss->title, doc, pcur->xmlChildrenNode);
              else if (!xml_strcmp (pcur->name, "description"))
                rss_read_copy (rss->desc, doc, pcur->xmlChildrenNode);
              else if (!xml_strcmp (pcur->name, "date") ||
                       !xml_strcmp (pcur->name, "pubDate") ||
                       !xml_strcmp (pcur->name, "dc:date"))
                rss->date = parse_date ((const char *) xml_get_string (pcur->xmlChildrenNode));

              pcur = xml_get_nextnode (pcur);
            }

        }
      else if ((!xml_strcmp (cur->name, "item")))
        {
          xml_node_t *pcur = xml_get_childnode (cur);
          st_rss_item_t *item = &rss->item[rss->item_count];
          int found = 0;
          char link[RSSMAXBUFSIZE], guid[RSSMAXBUFSIZE];

          *link = *guid = 0;

          while (pcur)
            {
              while (pcur && xml_is_empty_node (pcur))
                pcur = xml_get_nextnode (pcur);

              if (!pcur)
                break;

#ifdef  DEBUG
              printf ("%s\n", pcur->name);
              fflush (stdout);
#endif

              if (!xml_strcmp (pcur->name, "title"))
                {
                  rss_read_copy (item->title, doc, pcur->xmlChildrenNode);
                  found = 1;
                }
              else if (!xml_strcmp (pcur->name, "link"))
                {
                  rss_read_copy (link, doc, pcur->xmlChildrenNode);
                  found = 1;
                }
              else if (!xml_strcmp (pcur->name, "guid") && (!(*link)))
                {
                  rss_read_copy (guid, doc, pcur->xmlChildrenNode);
                  found = 1;
                }
              else if (!xml_strcmp (pcur->name, "description"))
                {
                  rss_read_copy (item->desc, doc, pcur->xmlChildrenNode);
                  found = 1;
                }
              else if (!xml_strcmp (pcur->name, "date") ||
                       !xml_strcmp (pcur->name, "pubDate") ||
                       !xml_strcmp (pcur->name, "dc:date"))
                { 
                  item->date = parse_date ((const char *) xml_get_string (pcur->xmlChildrenNode));
                  found = 1;
                }
#if 0
              else
                {
                  if (!found) // possibly malformed feed
                    break;
                  else
                    found = 0;
                }
#endif

              pcur = xml_get_nextnode (pcur);
            }

          // some feeds use the guid tag for the link
          if (*link)
            strcpy (item->url, link);
          else if (*guid)
            strcpy (item->url, guid);
          else
            *(item->url) = 0;

          rss->item_count++;

          if (rss->item_count == RSSMAXITEM)
            break;
        }

//      rss->item_count++;

      cur = xml_get_nextnode (cur);
    }

#ifdef  DEBUG
  rss_st_rss_t_sanity_check (rss);
  fflush (stdout);
#endif

  return rss;
}


static st_rss_t *
rss_open_atom (st_rss_t *rss)
{
  xml_doc_t *doc;
  xml_node_t *cur;

  doc = xml_parse (rss->url);
  if (!doc)
    {
      fprintf (stderr, "ERROR: cannot read %s\n", rss->url);
      return NULL;
    }

  cur = xml_get_rootnode (doc);
  if (!cur)
    {
      fprintf (stderr, "ERROR: empty document %s\n", rss->url);
      xml_free (doc);
      return NULL;
    }

  cur = cur->xmlChildrenNode;
  while (cur && xml_is_empty_node (cur))
    cur = cur->next;
  if (!cur)
    {
//      fprintf (stderr, "");
      return NULL;
    }

  while (cur)
    {
      while (cur && xml_is_empty_node (cur))
        cur = cur->next;

      if (!cur)
        break;

      if (!xml_strcmp (cur->name, "title"))
        rss_read_copy (rss->title, doc, cur->xmlChildrenNode);
      else if (!xml_strcmp (cur->name, "description"))
        rss_read_copy (rss->desc, doc, cur->xmlChildrenNode);
//      else if (!xml_strcmp (cur->name, "link"))
//        rss_read_copy (rss->url, doc, cur->xmlChildrenNode);
      else if (!xml_strcmp (cur->name, "date") ||
               !xml_strcmp (cur->name, "pubDate") ||
               !xml_strcmp (cur->name, "dc:date") ||
               !xml_strcmp (cur->name, "modified"))
        rss->date = parse_date ((const char *) xml_get_string (cur->xmlChildrenNode));
      else if ((!xml_strcmp (cur->name, "entry")))
        {
          xml_node_t *pcur = cur->xmlChildrenNode;
          st_rss_item_t *item = &rss->item[rss->item_count];
          int found = 0;

          while (pcur)
            {
              while (pcur && xml_is_empty_node (pcur))
                pcur = pcur->next;

              if (!pcur)
                break;

#ifdef  DEBUG
              printf ("%s\n", pcur->name);
              fflush (stdout);
#endif

              if (!xml_strcmp (pcur->name, "title"))
                {
                  rss_read_copy (item->title, doc, pcur->xmlChildrenNode);
                  found = 1;
                }
#if 1
              else if (!xml_strcmp (pcur->name, "id"))
                {
                  rss_read_copy (item->url, doc, pcur->xmlChildrenNode);
                  found = 1;
                }
#else
#if 0
<link rel="alternate" type="text/html" href="http://edition.cnn.com/2006/POLITICS/11/01/kerry.remarks/"/>
#endif
              else if (!xml_strcmp (pcur->name, "link"))
                {
                  while (pcur->properties)
                    {
                      printf ("%s\n\n", pcur->properties->name);
 
                      pcur->properties++;
                    }

                  found = 1;
                }
#endif
              else if (!xml_strcmp (pcur->name, "content"))
                {
                  rss_read_copy (item->desc, doc, pcur->xmlChildrenNode);
                  found = 1;
                }
              else if (!xml_strcmp (pcur->name, "modified"))
                { 
                  item->date = parse_date ((const char *) xml_get_string (pcur->xmlChildrenNode));
                  found = 1;
                }

              pcur = pcur->next;
            }

          rss->item_count++;

          if (rss->item_count == RSSMAXITEM)
            break;
        }

//      rss->item_count++;

      cur = cur->next;
    }

#ifdef  DEBUG
  rss_st_rss_t_sanity_check (rss);
  fflush (stdout);
#endif

  return rss;
}


st_rss_t *
rss_open (const char *fname)
{
  st_rss_t *rss = NULL;

  if (!(rss = malloc (sizeof (st_rss_t))))
    return NULL;

  memset (rss, 0, sizeof (st_rss_t));

  strncpy (rss->url, fname, RSSMAXBUFSIZE)[RSSMAXBUFSIZE - 1] = 0;
  rss->item_count = 0;

  rss->version = rss_demux (fname);

#ifdef  DEBUG
  fprintf (stderr, "version: %s\n", rss_get_version_s_by_id (rss->version));
  fflush (stderr);
#endif

  if (rss->version == -1)
    {
      fprintf (stderr, "ERROR: unknown feed format %s\n", rss->url);
      return NULL;
    }

  switch (rss->version)
    {
      case ATOM_V0_1:
      case ATOM_V0_2:
      case ATOM_V0_3:
        return rss_open_atom (rss);
      default:
        return rss_open_rss (rss);
    }

  free (rss);
  rss = NULL;

  return NULL;
}


int
rss_close (st_rss_t *rss)
{
  if (rss)
    {
      free (rss);
      rss = NULL;
    }

  return 0;
}


int
rss_write (FILE *fp, st_rss_t *rss, int version)
{
  unsigned int i = 0;

  if (!fp)
    return -1;

  if (!rss)
    return -1;

  if (version != 1) // default to RSS 2.0
    version = 2;

  fputs ("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", fp);

  if (version == 1)
    fputs ("<rdf:RDF xmlns=\"http://purl.org/rss/1.0/\">\n", fp);
  else
    fputs ("<rss version=\"2.0\">\n", fp);

  fputs ("  <channel>\n"
         "    <title>RSStool</title>\n"
         "    <link>http://rsstool.berlios.de</link>\n"
         "    <description>read, parse, merge and write RSS (and Atom) feeds</description>\n"
//         "    <dc:date>%ld</dc:date>"
         , fp);


  if (version == 1)
    {
      fputs ("<items>\n"
             "<rdf:Seq>\n", fp);

      for (i = 0; i < rss_item_count (rss); i++)
        fprintf (fp, "\n        <rdf:li rdf:resource=\"%s\"/>", rss->item[i].url);

      fputs ("</rdf:Seq>\n"
             "</items>\n"
             "</channel>\n", fp);
    }

  for (i = 0; i < rss_item_count (rss); i++)
    {
      if (version == 1)
        fprintf (fp, "<item rdf:about=\"%s\">\n", rss->item[i].url);
      else
        fputs ("    <item>\n", fp);

      fprintf (fp, "      <title>%s</title>\n", rss->item[i].title);
      fprintf (fp, "      <link>%s</link>\n", rss->item[i].url);
      fprintf (fp, "      <description>%s</description>\n", rss->item[i].desc);
      fprintf (fp, "      <dc:date>%ld</dc:date>\n", rss->item[i].date);

      fputs ("    </item>\n", fp);
    }

  if (version == 2)
    fputs ("  </channel>\n", fp);

  fputs ("</rss>\n", fp);

  return 0;
}
